!  ===================================================================72
!
!  File ENDSAM.F contains:
!    - module GLOB,
!    - module ENDF,
!    - module SAMPLING,
!    - main program ENDSAM.
!
!  AUTHOR
!
!    Lucijan Plevnik
!    Jozef Stefan Institute, Ljubljana, Slovenia, 2015
!
!  ACKNOWLEDGEMENT
!
!  The author would like to thank Gasper Zerovnik for starting the work
!  on this project, introducing the author to the topic and discussions.
!
!  REFERENCES
!
!  Routines and functions from libraries BLAS [1] and LAPACK [2] are
!  called from module SAMPLING. These libraries can be downloaded from
!  website [3]. Download archive lapack-3.5.0.tgz and follow
!  instructions in Readme file. Link to specific instructions for
!  WINDOWS installation can be found at [3].
!
!  [1] http://www.netlib.org/blas/faq.html#3
!  [2] http://www.netlib.org/lapack/
!  [3] http://www.netlib.org/lapack/#_software
!
!  Alternatively, folder min_lapack_blas contains the minimal set of
!  routines which are needed to compile endsam.f. Batch file may be used
!  for compilation.
!
!  =====================================================================
!  =====================================================================
      module glob
!  =====================================================================
!     Module GLOB contains global variables and routines.
!
!     [integer]
!       MAT,MF,MT,NS : line flags in ENDF file
!	    M2    : number of parameters per resonance in FILE2
!	    NR    : number of resonances in FILE2
!       NNC   : counter of nonconstant resonance parameters in FILE32
!       NZC   : number of nonzero nondiagonal correlations
!       UFILO : number of unit of output file
!	    IND32 : vector in which first NNC entries tell the position of
!               corresponding nonconstant parameter in 2-dimansional
!               vector MU. Positions in MU are counted in
!               colexicographic order.
!	    NONC  : vector of indices of nonconstant parameters
!    [integer(kind=1)]
!       TYPES : vector with with NNC entries which equal:
!            -1 : parameter is distributed minus lognormally
!             0 : parameter is distributed normally
!             1 : parameter is distributed lognormally
!    [double precision]
!       MU    : 2-dimensional vector of mean values of resonance
!               parameters. Second component contains index of
!               resonance, while the first one contains index of a
!               parameter of this resonance.
!   	MUNC  : vector of mean values of nonconstant parameters
!	    DEV   : vector of standard deviations of nonconstant parameters
!   	COR   : vector of lower triangle of correlation matrix of
!                nonconstant parameters (including diagonal) stored
!                column-wise
!    [logical]
!       WRI   : tells whether we write data to output files or not
!    [character]
!       COLS  : first 6 columns of length 11 in a record of ENDF file
!       OU,OV : are used to format numbers as strings
!
!  =====================================================================
      implicit none
      integer           :: m2,mat,mf,mt,nnc,nr,ns,nzc,ufilo
      logical           :: wri
      double precision  :: eps,one,zero
      character(len=3)  :: fora
      character(len=5)  :: fori
      character(len=8)  :: for6
      character(len=10) :: ou,ov
      character(len=17) :: fora6
      character(len=66) :: cols
      parameter(eps=1.0D-5,fora='(a)',fora6='(a66,i4,i2,i3,i5)',
     &for6='(6e11.0)',fori='(i10)',one=1.0D+0,zero=0.0D+0)
      integer,allocatable,dimension(:)            :: ind32,nonc
      integer(kind=1),allocatable,dimension(:)    :: types
      double precision,allocatable,dimension(:)   :: cor,dev,munc
      double precision,allocatable,dimension(:,:) :: mu
!  =====================================================================
      contains
!  =====================================================================
!  =====================================================================
      subroutine chext(inp,j,info)
!  =====================================================================
!
!  CHEXT prepends integer J to extension of filename INP.
!
!  =====================================================================
!  INPUT & OUTPUT ARGUMENTS
      character(len=*), intent(in)  :: inp
      integer,          intent(in)  :: j
      integer,          intent(out) :: info
!  LOCAL VARIABLES
      integer            :: i,k,l
      character(len=160) :: outt
!  =====================================================================
      info=0
      write(ou,fori) j
      k=len(trim(inp))
      do 10 i=k,1,-1
!
!  Find beginning of extension.
!
      if(inp(i:i).eq.'.') goto 20
10    continue
!
!  If no extension was found, then start with writing at the end of INP.
!
      i=k+1
20    outt(1:i-1)=inp(1:i-1)
      l=len(trim(adjustl(ou)))
!
!  Check if filename is not too long.
!
      if(k+l.lt.len(outt)) goto 25
      info=k+l+1-len(outt)
      return
!
!  Write filename+extension to OUTT.
!
25    outt(i:i+l)='_'//trim(adjustl(ou))
      outt(i+l+1:k+l+1)=inp(i:k)
      if(wri) goto 30
!
!  If WRI=.FALSE., then create a file with the name OUTT.
!
      open(j+9,file=outt(1:k+l+1),status='replace',action='write',
     &blank='zero')
      return
!
!  If WRI=.TRUE., then write data to the file OUTT.
!
30    write(j+9,fora6) cols,mat,mf,mt,ns
      return
!  =====================================================================
      end subroutine chext
!  =====================================================================
!  =====================================================================
      subroutine readl(uni,rec)
!  =====================================================================
!
!  READL reads a line from ENDF file.
!
!  =====================================================================
!  INPUT & OUTPUT ARGUMENTS
      integer,intent(in)    :: uni
      integer,intent(inout) :: rec
!  =====================================================================
      read(uni,fora6) cols,mat,mf,mt,ns
      rec=rec+1
      if(wri) write(ufilo,fora6) cols,mat,mf,mt,ns
      return
!  =====================================================================
      end subroutine readl
!  =====================================================================
!  =====================================================================
      subroutine cont(uni,l1,l2,n1,n2,rec)
!  =====================================================================
!
!  CONT reads a cont record from ENDF format and returns integer values
!  of flags.
!
!  =====================================================================
!  INPUT & OUTPUT ARGUMENTS
      integer,intent(in)    :: uni
      integer,intent(out)   :: l1,l2,n1,n2
      integer,intent(inout) :: rec
!  =====================================================================
      call readl(uni,rec)
      read(cols(23:66),10) l1,l2,n1,n2
      return
!  =====================================================================
!  FORMATS
10    format(4i11)
!  =====================================================================
      end subroutine cont
!  =====================================================================
!  =====================================================================
      subroutine listr(uni,l1,l2,n1,n2,rec)
!  =====================================================================
!
!  LISTR reads a list record from ENDF format and returns integer values
!  of flags.
!
!  =====================================================================
!  INPUT & OUTPUT ARGUMENTS
      integer,intent(in)    :: uni
      integer,intent(out)   :: l1,l2,n1,n2
      integer,intent(inout) :: rec
!  COUNTERS
      integer j
!  =====================================================================
      call cont(uni,l1,l2,n1,n2,rec)
      do 5 j=1,n2
      call readl(uni,rec)
5     continue
      return
!  =====================================================================
      end subroutine listr
!  =====================================================================
!  =====================================================================
      subroutine flaimpl(fla,flav,fm)
!  =====================================================================
!
!  FLAIMPL throws an error if a certain values of a flag is not
!  implemented yet.
!
!  =====================================================================
!  INPUT & OUTPUT ARGUMENTS
      character(*), intent(in) :: fla
      integer,      intent(in) :: flav,fm
!  =====================================================================
      write(ou,fori) flav
      write(ov,fori) fm
      write(*,fora) ' '
      write(*,fora) ' ERROR - value '//trim(adjustl(ou))//' of flag '//
     &trim(fla)//' in file '//trim(adjustl(ov))//
     &' is not implemented yet.'
      stop
!  =====================================================================
      end subroutine flaimpl
!  =====================================================================
!  =====================================================================
      subroutine invfla(fla,flav,fm)
!  =====================================================================
!
!  INVFLA throws an error if a flag has invalid value.
!
!  =====================================================================
!  INPUT & OUTPUT ARGUMENTS
      character(*), intent(in) :: fla
      integer,      intent(in) :: flav,fm
!  =====================================================================
      write(ou,fori) flav
      write(ov,fori) fm
      write(*,fora) ' '
      write(*,fora) ' ERROR - value '//trim(adjustl(ou))//' of flag '//
     &trim(fla)//' in file '//trim(adjustl(ov))//' is not valid.'
      stop
!  =====================================================================
      end subroutine invfla
!  =====================================================================
!  =====================================================================
      end module glob
!  =====================================================================
!  =====================================================================
      module endf
!  =====================================================================
!  Module ENDF contains variables and routines needed to read FILE32 in
!  ENDF file and save covaraince data for resonance parameters given
!  there.
!
!  VARIABLES
!    [integer]
!       JER  : counter of energy ranges in FILE32
!       NER  : number of energy ranges in FILE32
!	    NNN  : counter of resonance parameters with covariance data
!       NR2  : index of a certain resonance in FILE2
!       NR32 : counter of resonances in FILE32
!	    REC  : counter of records read from FILE32
!       UTEM : unit of temporary file where we store FILE32
!       LFW,LRF,LRU,NRO : flags in FILE32
!
!  =====================================================================
      use glob
      implicit none
      integer           :: jer,lfw,lrf,lru,ner,nnn,nr2,nr32,nro,rec,utem
      character(len=7)  :: fore
      character(len=11) :: emp,ze,zer
      data utem/9/
      parameter(emp='           ',fore='(e11.0)',ze=' 0.        ',
     &zer=' 0.000000+0')
!  =====================================================================
      contains
!  =====================================================================
!  =====================================================================
      subroutine file32(sav,info)
!  =====================================================================
!
!  FILE32 reads resonance parameters from file 32 in ENDF file.
!
!  INPUT & OUTPUT ARGUMENTS
!
!    [logical,in]
!       SAV :
!         = false: count resonance parameters with covariance data
!         = true:  save resonance parameters with covariance data
!
!    [integer,out]
!       INFO :
!          = 0: successful exit
!          > 0: if INFO=i+(j-1)*(2*nnn-j)/2 for i>j, then correlation
!               (i,j) is not between -1 and 1
!
!  =====================================================================
!  INPUT & OUTPUT ARGUMENTS
      logical, intent(in)  :: sav
      integer, intent(out) :: info
!  COUNTERS
      integer i,ii,j,jj,js,k,l
!  LOCAL VARIABLES
      integer isr,l1,l2,lcomp,m,mpar,n1,n2,nch,ndg,njs,njsx,nlrs,nls,nm,
     &nnc0,nnn1,nrb,nrow,nrs,nrsa,nsrs,nvs,sumj
      double precision                             :: a,factor
      double precision,allocatable,dimension(:)    :: temp,r,rr
      integer,allocatable,dimension(:)             :: ind
      integer(kind=1)                              :: tt
      integer(kind=1),allocatable,dimension(:)     :: ttemp
      character(len=11)                            :: s
      character(len=11),allocatable,dimension(:)   :: str
      character(len=11),allocatable,dimension(:,:) :: cov
!  =====================================================================
      if(sav) goto 5
!
!  Before counting covariance data give information about energy range.
!
      write(ou,fori) jer
      write(ov,fori) ner
      write(*,fora) ' '
      write(*,fora) ' Reading energy range '//trim(adjustl(ou))//
     &' of '//trim(adjustl(ov))//' in FILE32.'
      write(*,612) ' LRU=',lru,' LRF=',lrf
      goto 9
!
!  Before saving covariance data rewind temporary file which contains
!  FILE32 to the beginning of energy range records.
!
5     nnc0=nnc
      ind32=0
      do 8 i=1,rec
      backspace utem
8     continue
!
!  Initialization.
!
9     info=0
      nzc=0
      nr32=0
      nnc=0
      rec=0
!
!  Default value of NSRS is set to 1, but it may become different in the
!  case LRU=1,LCOMP=1.
!
      nsrs=1
      if(lru.eq.2) goto 127
!
!  LRU = 1
!
      if(nro.eq.1) call flaimpl('NRO',1,32)
      call cont(utem,l1,lcomp,njs,isr,rec)
      if(isr.eq.0) goto 12
      if(isr.eq.1) goto 10
      call invfla('ISR',isr,2)
10    if(lrf.eq.1 .or. lrf.eq.2) goto 11
      call listr(utem,l1,l2,n1,n2,rec)
      goto 12
11    call cont(utem,l1,l2,n1,n2,rec)
      
!
!  LCOMP may have values 0,1,2; the value 0 is possible only if
!  LRF = 1 or 2.
!
12    if(.not.sav) write(*,621) ' LCOMP=',lcomp
      if(lcomp.eq.2) goto 70
      if(lcomp.eq.1) goto 35
      if(lcomp.ne.0 .or. (lrf.ne.1 .and. lrf.ne.2)) call invfla('LCOMP',
     &lcomp,32)
!
!  LRU = 1 -> LCOMP = 0
!
      if(.not.sav) goto 14
      allocate(cov(nnn,nnn))
      cov=zer
14    nnn=0
      allocate(r(6),str(7))
      do 26 js=1,njs
      call cont(utem,l1,l2,n1,nrs,rec)
      do 24 i=1,nrs
!
!  Read resonance parameters and their covariances.
!
      read(utem,for6) (r(j),j=1,6)
      read(utem,611) (str(j),j=1,7)
      rec=rec+3
      if(.not.sav) goto 15
      cov(nnn+3,nnn+2)=str(3)
      cov(nnn+4,nnn+2)=str(5)
      cov(nnn+4,nnn+3)=str(6)
!
!  Find resonance from FILE32 in FILE2.
!
      call findres(6,r)
!
!  Count nonconstant parameters.
!  When SAV=.TRUE. save their means, deviations, and types.
!
15    do 22 j=1,6
      n2=0
      if(sav .and. nr2.ne.0) n2=m2*(nr2-1)+j
      tt=1
      goto (17,22,22,18,19,20), j
17    s=str(1)
      tt=0
      goto 21
18    s=str(2)
      goto 21
19    s=str(4)
      goto 21
20    s=str(7)
!
!  GF may have negative value and in this case it is distributed
!  according to minus lognormal distribution.
!
      if(r(6).lt.zero) tt=-1
21    call savec(sav,s,r(j),tt,n2,.true.)
22    continue
24    continue
26    continue
      deallocate(r,str)
      if(.not.sav) return
      nnn1=nnn
      goto 160
!
!  LRU = 1 -> LCOMP = 1
!
35    call cont(utem,l1,l2,nsrs,nlrs,rec)
      if(nsrs.lt.0) call invfla('NSRS',nsrs,32)
      if(nlrs.lt.0) goto 37
      if(nlrs.eq.0) goto 41
      if(lrf.eq.7) goto 37
      call flaimpl('NLRS',nlrs,32)
37    call invfla('NLRS',nlrs,32)
41    if(sav) allocate(temp(nnn),ttemp(nnn))
      nnn=0
!
!  If NSRS=0 and NLRS=0, then there are no resonance parameters with
!  covariance data. Therefore, ENDSAM will not call routine again with
!  the flag value SAV=.TRUE., so there is no need to deallocate vectors
!  TEMP and TTEMP.
!
      if(nsrs.eq.0) return
!
!  NNN1=number of resonance parameters with covariance data in current
!  short range section
!  Here, the loop over short range sections starts.
!
42    nnn1=0
      if(lrf.ne.7) goto 60
!
!  LRU = 1 -> LCOMP = 1 -> LRF = 7
!
      call cont(utem,njsx,l2,n1,n2,rec)
      do 50 js=1,njsx
      call cont(utem,nch,nrb,n1,n2,rec)
      if(nch.le.0 .or. nch.ge.m2) call invfla('NCH',nch,32)
      allocate(r(1+nch))
      do 45 i=1,nrb
!
!  Read resonance parameters with covariance data.
!
      read(utem,for6) (r(j),j=1,1+nch)
      rec=rec+1+nch/6
!
!  Find resonance from FILE32 in FILE2.
!
      if(sav) call findres(1+nch,r)
      do 44 j=1,1+nch
      nnn1=nnn1+1
      nnn=nnn+1
      if(.not.sav) goto 44
!
!  Save resonance parameters, their types and indices in vector MU into
!  vectors TEMP, TTEMP and IND32, respectively.
!
      if(nr2.ne.0) ind32(nnn)=m2*(nr2-1)+j
      temp(nnn)=r(j)
      if(j.eq.1) goto 43
      ttemp(nnn)=1
      goto 44
43    ttemp(nnn)=0
44    continue
45    continue
      deallocate(r)
50    continue
!
!  Read cont record which contains size of covariance matrix.
!
      call cont(utem,l1,l2,nvs,n2,rec)
      if(n2.ne.nnn1) call invfla('NPARB',n2,32)
      if(2*nvs.ne.nnn1*(nnn1+1)) call invfla('N',nvs,32)
      goto 140
!
!  LRU = 1 -> LCOMP = 1 -> LRF = 1 or 2 or 3
!
60    call cont(utem,mpar,l2,n1,nrb,rec)
      if(mpar.lt.1 .or. mpar.gt.5) call invfla('MPAR',mpar,32)
      allocate(r(6))
      do 67 i=1,nrb
!
!  Read resonance parameters with covariance data.
!
      read(utem,for6) (r(j),j=1,6)
      rec=rec+1
!
!  Find resonance from FILE32 in FILE2.
!
      if(sav) call findres(6,r)
!
!  Save resonance parameters, their types and indices in vector MU into
!  vectors TEMP, TTEMP and IND32, respectively.
!
      do 65 j=1,6
      if(j.eq.2 .or. j.gt.mpar+1) goto 65
      nnn1=nnn1+1
      nnn=nnn+1
      if(.not.sav) goto 65
      if(nr2.ne.0) ind32(nnn)=m2*(nr2-1)+j
      temp(nnn)=r(j)
      ttemp(nnn)=1
      if(j.lt.5 .or. (j.lt.6 .and. lrf.ne.3)) goto 63
!
!  GF may have negative value and in this case it is distributed
!  according to minus lognormal distribution.
!
      if(r(j).lt.zero) ttemp(nnn)=-1
      goto 65
63    if(j.eq.1) ttemp(nnn)=0
65    continue
67    continue
      deallocate(r)
      nvs=nnn1*(nnn1+1)/2
      goto 140
!
!  LRU = 1 -> LCOMP = 2
!
70    if(lrf.ne.7) goto 100
!
!  LRU = 1 -> LCOMP = 2 -> LRF = 7
!
      call listr(utem,l1,njsx,n1,n2,rec)
      if(njsx.lt.0 .or. njsx.gt.njs) call invfla('NJSX',njsx,32)
      nnn=0
      do 90 js=1,njs
      call listr(utem,l1,l2,n1,nch,rec)
      if(nch.le.0) call invfla('NCH',nch,32)
      call cont(utem,l1,nrsa,n1,n2,rec)
      allocate(r(1+nch),str(1+nch))
      do 89 i=1,nrsa
!
!  Read resonance parameters with covariance data and their deviations.
!
      read(utem,for6) (r(j),j=1,1+nch)
      read(utem,611) (str(j),j=1,1+nch)
      rec=rec+2*(1+nch/6)
!
!  Find resonance from FILE32 in FILE2.
!
      if(sav) call findres(1+nch,r)
!
!  Count nonconstant parameters.
!  When SAV=.TRUE. save their means, deviations, and types.
!
      do 88 j=1,1+nch
      if(.not.sav) goto 86
      n2=0
      if(nr2.ne.0) n2=m2*(nr2-1)+j
      if(j.gt.1) goto 84
      tt=0
      goto 86
84    tt=1
86    call savec(sav,str(j),r(j),tt,n2,.false.)
88    continue
89    continue
      deallocate(r,str)
90    continue
!
!  NM=number of lines with data about correlations
!  NDG=number of digits for the correlation matrix
!
      call cont(utem,ndg,nnn,nm,n2,rec)
      goto 113
!
!  LRU = 1 -> LCOMP = 2 -> LRF = 1 or 2 or 3
!
100   call cont(utem,l1,l2,n1,nrsa,rec)
      if(nrsa.gt.0) goto 101
      if(nrsa.lt.0) call invfla('NRSA',nrsa,32)
      nnn=0
      return
101   allocate(r(6*nrsa),str(6*nrsa))
!
!  K = counter of means and deviations of resonance parameters
!
      k=0
      do 103 i=1,nrsa
!
!  Read resonance parameters with covariance data and their deviations.
!
      read(utem,for6) (r(k+j),j=1,6)
      read(utem,611) (str(k+j),j=1,6)
      k=k+6
      rec=rec+2
103   continue
      call cont(utem,ndg,nnn,nm,n2,rec)
!
!  M=number of resonance parameters with covariance data per line
!
      m=4
      if(lrf.eq.3) m=5
!
!  L = NNN/NRSA; L may be equal 3 or 4 or 5 and cannot be larger than M
!
      l=m+1
      if(nnn.eq.5*nrsa) l=5
      if(nnn.eq.4*nrsa) l=4
      if(nnn.eq.3*nrsa) l=3
      if(l.gt.m) call invfla('NNN',nnn,32)
      k=0
      nnn=0
      do 111 i=1,nrsa
!
!  Find resonance from FILE32 in FILE2.
!
      if(sav) call findres(6,r(k+1:k+6))
!
!  Eliminate those resonance parameters which are not included in
!  correlation matrix.
!  Count nonconstant parameters.
!  When SAV=.TRUE. save their means, deviations, and types.
!
      do 110 j=1,6
      k=k+1
      if(j.gt.6-m+l) goto 110
      n2=0
      if(sav .and. nr2.ne.0) n2=m2*(nr2-1)+j
      tt=1
      goto (105,110,106,109,107,108), j
105   tt=0
      goto 109
106   if(lrf.ne.3) goto 110
      goto 109
107   if(lrf.ne.3) goto 109
!
!  GF may have negative value and in this case it is distributed
!  according to minus lognormal distribution.
!
108   if(r(k).lt.zero) tt=-1
109   call savec(sav,str(k),r(k),tt,n2,.false.)
110   continue
111   continue
      deallocate(r,str)
!
!  LRU = 1 -> LCOMP = 2
!
113   if(.not.sav .or. nm.eq.0) return
      if(nm.lt.0) call invfla('NM',nm,32)
      if(ndg.lt.2 .or. ndg.gt.6) call invfla('NDIGIT',ndg,32)
      if(ndg.eq.2) nrow=18
      if(ndg.eq.3) nrow=13
      if(ndg.eq.4) nrow=11
      if(ndg.eq.5) nrow=9
      if(ndg.eq.6) nrow=8
      factor=dble(10**ndg)
      allocate(ind(nrow))
      do 126 j=1,nm
!
!  Read the INTG record.
!
      if(ndg.eq.2) read(utem,502) ii,jj,(ind(i),i=1,nrow)
      if(ndg.eq.3) read(utem,503) ii,jj,(ind(i),i=1,nrow)
      if(ndg.eq.4) read(utem,504) ii,jj,(ind(i),i=1,nrow)
      if(ndg.eq.5) read(utem,505) ii,jj,(ind(i),i=1,nrow)
      if(ndg.eq.6) read(utem,506) ii,jj,(ind(i),i=1,nrow)
      rec=rec+1
      if(ii.gt.nnn) goto 121
!
!  Find index K<=NNC such that NONC(K)=II.
!
      call findind(ii,jj,k,nnc,l1)
      if(l1.ne.0) goto 126
!
!  Interpret the INTG record and save correlations.
!
121   jj=jj-1
      do 125 i=1,nrow
      jj=jj+1
      if(ind(i).eq.0) goto 125
      if(ii.le.nnn) goto 122
      write(ou,fori) nnn
      write(ov,fori) ii
      write(s(1:10),fori) jj
      write(*,fora) ' '
      write(*,fora) ' WARNING: Correlation matrix is of size '//
     &trim(adjustl(ou))//', so it cannot contain nonzero entry at posit'
     &//'ion ('//trim(adjustl(ov))//','//trim(adjustl(s(1:10)))//').'
      goto 125
122   if(jj.lt.ii) goto 123
      write(ou,fori) ii
      write(ov,fori) jj
      write(*,fora) ' WARNING: Strictly lower triangle of correlation m'
     &//'atrix must be given, so there cannot be nonzero entry at posit'
     &//'ion ('//trim(adjustl(ou))//','//trim(adjustl(ov))//').'
      goto 125
!
!  Find index L<K such that NONC(L)=JJ.
!
123   call findind(jj,ii,l,k-1,l1)
      if(l1.ne.0) goto 125
      nzc=nzc+1
      a=dble(ind(i))
      cor(k+(l-1)*(2*nnc-l)/2)=(a+sign(5.0D-1,a))/factor
125   continue
126   continue
      deallocate(ind)
      return
!
!  LRU = 2
!
!  K = counter of all resonance parameters
!
127   k=0
      call cont(utem,l1,l2,nls,n2,rec)
      if(.not.sav) goto 128
      allocate(temp(5*nnn),ttemp(5*nnn),ind(6))
      ind=(/1,0,3,4,5,2/)
      ttemp=1
128   allocate(r(6))
!
!  SUMJ = sum of the values NJS for each L
!
      sumj=0
      do 132 i=1,nls
      call cont(utem,l1,l2,n1,njs,rec)
      sumj=sumj+njs
!
!  Read resonance parameters D,AJ,GNO,GG,GF,GX.
!
      do 131 js=1,njs
      read(utem,for6) (r(j),j=1,6)
      rec=rec+1
      if(.not.sav) goto 131
      allocate(rr(5))
      rr=(/r(1),r(6),r(3),r(4),r(5)/)
!
!  Find resonance from FILE32 in FILE2.
!
      call findres(5,rr)
      do 130 j=1,6
      if(j.eq.2) goto 130
      k=k+1
      if(nr2.ne.0) ind32(k)=m2*(nr2-1)+ind(j)+1
      temp(k)=rr(ind(j))
!
!  GF may have negative value and in this case it is distributed
!  according to minus lognormal distribution.
!
      if(j.eq.5 .and. temp(k).lt.zero) ttemp(k)=-1
130   continue
      deallocate(rr)
131   continue
132   continue
      deallocate(r)
      if(sav) deallocate(ind)
      call cont(utem,mpar,l2,nvs,nnn,rec)
      if(mpar.lt.1 .or. mpar.gt.5) call invfla('MPAR',mpar,32)
      if(nnn.ne.mpar*sumj) call invfla('NPAR',nnn,32)
      if(2*nvs.ne.nnn*(nnn+1)) call invfla('NPAR*(NPAR+1)/2',nvs,32)
      if(.not.sav) goto 138
!
!  Eliminate those resonance parameters which are not included in
!  covariance matrix.
!
      k=0
      nnn=0
      do 135 i=1,sumj
      do 133 j=1,mpar
      nnn=nnn+1
      k=k+1
!
!  If LFW=0 and MPAR = 4, then resonance parameters are D,GNO,GG,GX.
!
      if(mpar.eq.4 .and. j.eq.4 .and. lfw.eq.0) k=k+1
      ind32(nnn)=ind32(k)
      temp(nnn)=temp(k)
      ttemp(nnn)=ttemp(k)
133   continue
      if(mpar.ne.4 .or. lfw.ne.0) k=k+5-mpar
135   continue
138   nnn1=nnn
!
!  Read lower triangle of covariance matrix.
!
140   allocate(cov(nnn1,nnn1))
      read(utem,611) (cov(j:nnn1,j),j=1,nnn1)
      rec=rec+1+(nvs-1)/6
      nnn=nnn-nnn1
!
!  Count nonconstant parameters.
!  When SAV=.TRUE. save their means, deviations, and types.
!
      do 152 j=1,nnn1
      k=nnn+1
      if(.not.sav) goto 145
      a=temp(k)
      tt=ttemp(k)
      l1=ind32(k)
145   call savec(sav,cov(j,j),a,tt,l1,.true.)
152   continue
      if(sav .and. nsrs.eq.1) deallocate(temp,ttemp)
160   nsrs=nsrs-1
      if(.not.sav) goto 165
!
!  Transform covariances of current NSRS block into correlations.
!
      l=nnn-nnn1
      k=0
      do 164 j=1,nnc0
      do 163 i=j,nnc0
      k=k+1
      if(i.eq.j .or. nonc(j).le.l .or. i.gt.nnc) goto 163
      s=cov(nonc(i)-l,nonc(j)-l)
      if(s.eq.zer .or. s.eq.emp .or. s.eq.ze) goto 163
      read(s,fore) a
      if(a.eq.zero) goto 163
!
!  Count nonzero nondiagonal correlations.
!
      nzc=nzc+1
      a=a/(dev(i)*dev(j))
!
!  If LRU=2, then relative covariances are given in ENDF file.
!
      if(lru.eq.2) a=a*munc(i)*munc(j)
!
!  Correlations must have values between -1 and 1 with tolerance EPS.
!
      if(abs(a).gt.one+eps) goto 222
      if(abs(a).gt.one) a=sign(one,a)
      cor(k)=a
163   continue
164   continue
165   deallocate(cov)
      if(nsrs.gt.0) goto 42
      return
!  =====================================================================
!  ERRORS
222   deallocate(cov)
      if(nsrs.gt.0) deallocate(temp,ttemp)
      info=nonc(i)+(nonc(j)-1)*(2*nnn-nonc(j))/2
      return
!  =====================================================================
!  FORMATS
502   format(2i5,1x,18i3)
503   format(2i5,1x,13i4)
504   format(2i5,1x,11i5)
505   format(2i5,1x,9i6)
506   format(2i5,8i7)
611   format(6a11)
612   format(2(a5,i1))
621   format(a7,i1)
!  =====================================================================
      end subroutine file32
!  =====================================================================
!  =====================================================================
      subroutine savec(sav,s,a,tt,i32,v)
!  =====================================================================
!
!  SAVEC saves covariance data into corresponding vectors.
!
!  [LOGICAL]
!     SAV : as in routine FILE32
!     V :
!       = T : S stands for variance
!       = F : S stands for deviation
!     A : mean
!     TT : type
!     I32 : index in vector IND32
!
!  =====================================================================
!  INPUT & OUTPUT ARGUMENTS
      logical,           intent(in)    :: sav,v
      character(len=11), intent(inout) :: s
      double precision,  intent(in)    :: a
      integer(kind=1),   intent(in)    :: tt
      integer,           intent(in)    :: i32
!  LOCAL VARIABLES
      double precision :: b
!  =====================================================================
      nnn=nnn+1
      if(s(1:1).ne.'-' .or. .not.sav) goto 10
!
!  If deviation or variance is negative, then change its sign and throw
!  a warning.
!
      write(*,fora) ' '
      write(ou,fori) nnn
      if(v) goto 6
      ov=' deviation'
      goto 8
6     ov=' variance '
8     write(*,fora) ' WARNING: parameter '//trim(adjustl(ou))//
     &' has negative'//trim(ov)//'.'
      s(1:1)=' '
10    if(s.eq.zer .or. s.eq.emp .or. s.eq.ze) return
      read(s,fore) b
      if(b.eq.zero) return
!
!  Count nonconstant parameters. If SAV=.TRUE., then save their data.
!
      nnc=nnc+1
      if(.not.sav) return
      nonc(nnc)=nnn
      munc(nnc)=a
      dev(nnc)=b
      if(v) dev(nnc)=sqrt(dev(nnc))
!
!  If LRU=2, then realative variance is given in ENDF file.
!
      if(lru.eq.2) dev(nnc)=dev(nnc)*a
      types(nnc)=tt
      ind32(nnc)=i32
      return
!  =====================================================================
      end subroutine savec
!  =====================================================================
!  =====================================================================
      subroutine findind(i,j,l,max,info)
!  =====================================================================
!
!  FINDIND finds index L<=MAX of integer I in increasingly ordered
!  vector NONC.
!
!  INPUT & OUTPUT ARGUMENTS
!
!  [integer,in]
!    I : index of parameter which we are looking for in vector NONC
!    J : index of parameter which is correlated to I
!  [integer,out]
!    L    : NONC(L)=I if INFO=0
!    INFO :
!      =0: successful exit
!      >0: I was not found in vector NONC
!
!  =====================================================================
!  INPUT & OUTPUT ARGUMENTS
      integer, intent(in)    :: i,j,max
      integer, intent(out)   :: l,info
!  =====================================================================
      info=0
      do 10 l=1,max
      if(nonc(l).eq.i) return
      if(nonc(l).gt.i) goto 20
10    continue
20    info=l
      write(ou,fori) i
      write(ov,fori) j
      write(*,fora) ' '
      write(*,fora) ' WARNING: parameter '//trim(adjustl(ou))//' is con'
     &//'stant, but it has nonzero correlation with parameter '
     &//trim(adjustl(ov))//'.'
      return
!  =====================================================================
      end subroutine findind
!  =====================================================================
!  =====================================================================
      subroutine findres(m,r)
!  =====================================================================
!
!  FINDRES finds index of resonance from FILE32 in FILE2.
!
!  INPUT & OUTPUT ARGUMENTS
!
!  [integer,in]
!    M : length of vector R
!  [double precision,inout]
!    R : vector of means of parameters of resonance
!
!  =====================================================================
!  INPUT & OUTPUT ARGUMENTS
      integer,          intent(in)    :: m
      double precision, intent(inout) :: r(m)
!  LOCAL VARIABLES
      integer          :: i,j,k
      double precision :: a
!  =====================================================================
      nr32=nr32+1
      do 15 j=1,nr
      do 10 i=1,m
      if(lru.eq.2) goto 5
      k=i
      goto 7
5     k=i+1
7     if(mu(k,j).eq.r(i)) goto 10
      a=abs(mu(k,j)-r(i))/max(abs(r(i)),abs(mu(k,j)))
!
!  For the first parameter allow relative tolerance EPS, while for the
!  other parameters allow relative tolerance 10%.
!
      if(a.gt.eps .and. (i.eq.1 .or. a.gt.1.0D-1)) goto 15
!
!  If resonance parameter from FILE32 was found in FILE2 with prescribed
!  relative tolerance, then work with the value from FILE2.
!
      r(i)=mu(k,j)
10    continue
      nr2=j
      return
15    continue
!
!  If resonance was not found in FILE2, then set its index to be 0.
!
      nr2=0
      write(ou,fori) nr32
      write(ov,fori) jer
      write(*,fora) ' WARNING: resonance '//trim(adjustl(ou))//' from e'
     &//'nergy range '//trim(adjustl(ov))//' in FILE32 not found in FILE
     &2.'
      return
!  =====================================================================
      end subroutine findres
!  =====================================================================
!  =====================================================================
      end module endf
!  =====================================================================
!  =====================================================================
      module sampling
!  =====================================================================
!  Module SAMPLING contains variables and routines needed to make random
!  samples of resonance parameters and write them to FILE2 in ENDF-6
!  format.
!
!  VARIABLES
!    [integer]
!        N   : Number of samples.
!	    REC2 : counter of records read from FILE2
!       UFIL : unit of ENDF file
!       LRF2,LRU2,NRO2 : flags in FILE2
!       LOGN(MAX(1,NNC*(NNC-1)/2)) : in LOG_NOR_SAMPLE or TRANSFORM, if
!             INFO>0, then first INFO entries equal I+(2*NNC-J)*(J-1)/2;
!             0<J<I<=NNC, if correlation (I,J) does not satisfy
!             theoretical bounds for (minus) lognormal distribution.
!    [logical]
!       MULNR(NNC) : vector whose entries tell whether nonconstant
!                    parameter has valid mean value or not
!    [double precision]
!        MU1(NNC)  : mean vector of normal parameters and logarithms of
!                    absolute values of +-lognormal parameters
!        DEV1(NNC) : vector of deviations of normal parameters and
!                    logarithms of absolute values of +-lognormal
!                    parameters
!       C(NNC*(NNC+1)/2) : If INFO=0 in subroutine TRANSFORM, then C
!                          contains lower triangle of correlation matrix
!                          of normal parameters and logarithms of
!                          absolute values of +-lognormal parameters.
!                          If INFO>0, then C is not referenced in
!                          TRANSFORM.
!
!  REFERENCES
!
!  Routines in this module call routines and functions from libraries
!  BLAS [1] and LAPACK [2] which can be downloaded from website [3].
!  Download archive lapack-3.5.0.tgz and follow instructions in Readme
!  file. Link to specific instructions for WINDOWS installation can be
!  found at [3].
!
!  [1] http://www.netlib.org/blas/faq.html#3
!  [2] http://www.netlib.org/lapack/
!  [3] http://www.netlib.org/lapack/#_software
!
!  =====================================================================
      use glob
      implicit none
      integer :: lrf2,lru2,n,nro2,rec2,ufil
      integer,allocatable,dimension(:)          :: logn
      logical,allocatable,dimension(:)          :: mulnr
      double precision,allocatable,dimension(:) :: mu1,dev1,c
      data ufil/2/
!  =====================================================================
      contains
!  =====================================================================
!  =====================================================================
      subroutine log_nor_sample(info)
!  =====================================================================
!
!  LOG_NOR_SAMPLE generates samples of random parameters with given mean
!  values, standard deviations, and correlations. Some of parameters are
!  distributed normally, some are distributed lognormally, while others
!  are distributed minus lognormally.
!
!  INPUT & OUTPUT ARGUMENTS
!
!    [out,integer]
!        INFO :
!          =   0   : successful exit,
!          = I > 0 : there are I pairs of parameters such that their
!                    correlations do not satisfy theoretical bounds
!                    for (minus) lognormal distribution.
!
!  DESCRIPTION
!
!  Subroutine works in the following steps.
!  1) Check if original correlation matrix is positive definite.
!  2) Call routine TRANSFORM.
!  3) At this point, we have mean and deviation vector and correlation
!     matrix C for a collection of normal parameters. We check whether
!     this correlation matrix is positive definite by trying to
!     compute its Cholesky factorization.
!  3a) If matrix is positive definite, then we compute normal samples
!      by formula
!                         mu1(i)+dev1(i)*L*y,  (*)
!      where i is index of parameter, y is Gaussian random vector, and
!      L is a lower triangular cholesky factor of C.
!  3b) Otherwise, we compute nearest (in Frobenius norm) correlation
!      matrix to C and then apply 3a).
!  4) We apply exponential function to samples of logarithms of
!     lognormal parameters. We change sign of samples of minus
!     lognormal parameters.
!
!  ROUTINES AND FUNCTIONS FROM LIBRARIES LAPACK AND BLAS:
!    DPPTRF, DLARNV, DTPMV
!
!  =====================================================================
!  INPUT & OUTPUT ARGUMENTS
      integer, intent(out) :: info
!  COUNTERS
      integer              :: i,j,k,l
!  LOCAL VARIABLES
      integer              :: info1,iseed(4),np
      real(kind=4)         :: finish,start
      character(len=1)     :: uplo
      double precision     :: sam(nnc),sei
      data iseed(1),iseed(2),iseed(3),iseed(4)/1,2341,3999,436/
      double precision,allocatable,dimension(:) :: c1
!  =====================================================================
      if(nzc.eq.0) goto 5
      uplo='l'
      allocate(c(nnc*(nnc+1)/2))
!
!  If correlation matrix is not identity matrix, then keep original
!  matrix in vector COR and work with its copy C.
!
      c=cor
!
!  Check if original correlation matrix is positive definite by
!  trying to compute its Cholesky decomposition.
!
      call dpptrf(uplo,nnc,c,info1)
      if(info1.eq.0) goto 5
      write(*,fora) ' '
      write(*,fora) ' WARNING: original correlation matrix is not positi
     &ve definite.'
!
!  Compute means,deviations, and correlations of absolute values of
!  logarithms of +-lognormal parameters.
!
5     allocate(mu1(nnc),dev1(nnc),mulnr(nnc))
      call transform(info)
!
!  If correlation matrix equals identity matrix, start making samples.
!
      if(nzc.eq.0) goto 27
!
!  If:
!  - either all parameters are normal,
!  - or correlation matrix is not compatible with +-lognormal
!    distribution,
!  then:
!  (a) if original correlation matrix is positive definite, then use
!      its Cholesky decomposition,
!  (b) otherwise, compute Cholesky decomposition of nearest correlation
!      matrix.
!  Otherwise, check if transformed correlation matrix is positive
!  definite by trying to compute its Cholesky decomposition and then
!  apply (a) or/and (b).
!
      do 15 j=1,nnc
      if(types(j).ne.0) goto 18
15    continue
      goto 20
18    if(info.eq.0) goto 25
20    if(info1.eq.0) goto 27
      c=cor
      goto 26
25    allocate(c1(nnc*(nnc+1)/2))
      c1=c
      call dpptrf(uplo,nnc,c,info1)
      if(info1.gt.0) c=c1
      deallocate(c1)
      if(info1.eq.0) goto 27
26    write(*,fora) ' '
      write(*,fora) ' WARNING: transformed correlation matrix is not pos
     &itive definite.'
      call cpu_time(start)
      call ncormat(np,sei,info1)
      call cpu_time(finish)
      if(info1.eq.-1) stop ' NCORMAT: iteration failed to converge.'
      write(ou,fori) np
      write(*,fora) ' It has '//trim(adjustl(ou))//' nonpositive eigenva
     &lue(s).'
      write(*,300) sei
      write(*,350) finish-start
      goto 25
27    do 50 j=1,n
!
!  Create a random vector of standard normal parameters and save into
!  vecotr SAM.
!
      call dlarnv(3,iseed,nnc,sam)
!
!  Apply Cholesky factor on random vector.
!
      if(nzc.gt.0) call dtpmv(uplo,'n','n',nnc,c,sam,1)
      do 40 i=1,nnc
!
!  If parameter has invalid mean value, then don't compute its sample.
!
      if(.not.mulnr(i)) goto 40
      sam(i)=mu1(i)+dev1(i)*sam(i)
!
!  Apply exponent on lognormal parameters.
!
      if(types(i).eq.0) goto 30
      sam(i)=exp(sam(i))
      if(types(i).lt.0) sam(i)=-sam(i)
!
!  Save sampled value into correct position vector MU.
!
30    info1=ind32(i)
      if(info1.eq.0) goto 40
      do 38 l=1,nr
      do 34 k=1,m2
      info1=info1-1
      if(info1.gt.0) goto 34
      mu(k,l)=sam(i)
      goto 40
34    continue
38    continue
40    continue
!
!  Write sampled vector MU into FILE2.
!
      ufilo=j+9
      call file2(.true.)
50    continue
      deallocate(mu1,dev1,mulnr)
      if(nzc.gt.0) deallocate(c)
      return
!  =====================================================================
!  FORMATS
300   format(' The smallest one equals ',es13.6,'.')
350   format(' Approximation for nearest correlation matrix was ',
     &'computed in ',es9.3,' seconds.')
!  =====================================================================
      end subroutine log_nor_sample
!  =====================================================================
!  =====================================================================
      subroutine transform(info)
!  =====================================================================
!
!  TRANSFORM gets reads mean values, deviations, and correlations
!  of a collection of normal, lognormal and minus lognormal random
!  parameters from vectors MUNC, DEV, and COR, respectively. It
!  transforms respectively means, deviations, and correlations of
!  lognormal and minus lognormal parameters to data of logarithms of
!  their absolute values and saves them into vectors MU1, DEV1, and
!  COR1, respectively.
!  Lower triangle of correlation matrix must be stored column-wise in
!  packed form.
!
!  Subroutine works in the following steps.
!  1) Check is mean values if parameters are consistent with their
!     types, that is:
!     - lognormal parameters must have positive mean,
!     - minus lognormal parameters must have negative mean.
!     For +-lognormal parameters X, compute means and deviations of
!     logarithms of their absolute values by formulas:
!     dev(log(|X|)) = sqrt(log(1+ru(X)**2)),
!     mu(log(|X|)) = log(|mu(X)|)-log(1+ru(X)**2)/2,
!     where ru stands for relative uncertainty.
!  2) Check which correlations between +-lognormal parameters satisfy
!     theoretical bounds. Namely, if X is +-lognormal, then we must
!     have:
!      for +-lognormal Y:
!            1+cor(X,Y)*ru(X)*ru(Y) <= 
!                   exp(sqrt(log(1+ru(X)**2)*log(1+ru(Y)**2)))
!      and
!            1+cor(X,Y)*ru(X)*ru(Y) >=
!                  exp(-sqrt(log(1+ru(X)**2)*log(1+ru(Y)**2))),
!      for normal Y:
!            abs(cor(X,Y)) <= sqrt(log(1+ru(X)**2))/abs(ru(X)).
!  2a) If these conditions are satisfied for all X and Y, then compute
!      correlations of logarithms of absolute values of +-lognormal
!      parameters X, that is:
!        for +-lognormal Y:
!          log(1+cor(X,Y)*ru(X)*ru(Y))/(dev(log(|X|))*dev(log(|Y|))),
!        for normal Y:
!                        cor(X,Y)*ru(X)/dev(log(|X|)).
!      Save correlations into C.
!  2b) Otherwise leave C unchanged.
!
!  INPUT & OUTPUT ARGUMENTS
!
!    [out,integer]
!        INFO    :
!          =   0   : successful exit,
!          = I > 0 : there are I pairs of parameters such that their
!                    correlations do not satisfy theoretical bounds
!                    for lognormal or minus lognormal distribution.
!
!  =====================================================================
!  INPUT & OUTPUT ARGUMENTS
      integer, intent(out) :: info
!  COUNTERS
      integer              :: i,j,k
!  LOCAL VARIABLES
      double precision     :: a,b
!  =====================================================================
      info=0
!
!  Check if +-lognormal parameters have the correct signs of mean value.
!  If they do, then compute means and deviations of logarithms of their
!  absolute values. If j-th parameter is inconsistent, then save .FALSE.
!  to j-th entry of vector MULNR and change type of parameter into
!  normal.
!
      k=0
      do 8 j=1,nnc
      mu1(j)=munc(j)
      dev1(j)=dev(j)
      mulnr(j)=.true.
      if(types(j).eq.0) goto 8
      if(dble(types(j))*mu1(j).gt.zero) goto 5
      mulnr(j)=.false.
      types(j)=0
      write(ou,fori) nonc(j)
      write(*,fora) ' WARNING: parameter '//trim(adjustl(ou))//', is eit
     &her lognormal with nonpositive mean or minus lognormal with nonneg
     &ative mean. It will not be sampled.'
      goto 8
5     mu1(j)=abs(mu1(j))
      dev1(j)=dev1(j)/mu1(j)
      a=log(one+dev1(j)**2)
!
!  If dev1(j)**2 is so small that log(1+dev1(j)**2) is computed to be 0,
!  then use approximation log(1+dev1(j)**2)=dev1(j)**2.
!
      if(a.gt.zero) goto 6
      mu1(j)=log(mu1(j))-dev1(j)**2/2
      goto 8
6     dev1(j)=sqrt(a)
      mu1(j)=log(mu1(j))-a/2
8     continue
!
!  If COR equals identity matrix, then leave C unchanged.
!
      if(nzc.eq.0) return
!
!  For every correlation of +-lognormal parameter and another parameter
!  check if correlation of logarithm of absolute value of +-lognormal
!  parameter is between -1 and 1 with tolerance EPS.
!
      k=0
      do 22 j=1,nnc
      do 20 i=j,nnc
      k=k+1
!
!  Skip diagonal entries and pairs of two normal parameters.
!
      if(i.eq.j .or. (types(i).eq.0 .and. types(j).eq.0)) goto 20
      a=cor(k)
      if(types(j).ne.0) goto 10
!
!  If j-th parameter is noraml, then i-th must be +-lognormal.
!
      a=a*dev(i)/(munc(i)*dev1(i))
      goto 12
10    a=a*dev(j)/munc(j)
      if(types(i).ne.0) goto 14
      a=a/dev1(j)
12    if(abs(a).gt.one+eps) goto 19
      goto 20
14    a=one+a*dev(i)/munc(i)
      b=exp((one+eps)*dev1(j)*dev1(i))
      if(a.gt.b .or. a.lt.one/b) goto 19
      goto 20
19    info=info+1
      logn(info)=k
20    continue
22    continue
      if(info.gt.0) return
!
!  If correlations and types are consistent, then save correlations of
!  logarithms of absolute values of +-lognormal parameters into C.
!
      k=0
      do 42 j=1,nnc
      do 40 i=j,nnc
      k=k+1
      c(k)=cor(k)
      if(i.eq.j .or. (types(i).eq.0 .and. types(j).eq.0)) goto 40
      if(types(j).ne.0) goto 25
      c(k)=c(k)*dev(i)/(munc(i)*dev1(i))
      goto 35
25    c(k)=c(k)*dev(j)/munc(j)
      if(types(i).ne.0) goto 30
      c(k)=c(k)/dev1(j)
      goto 35
30    c(k)=log(one+c(k)*dev(i)/munc(i))/(dev1(j)*dev1(i))
35    if(abs(c(k)).gt.one) c(k)=sign(one,c(k))
40    continue
42    continue
      return
!  =====================================================================
      end subroutine transform
!  =====================================================================
!  =====================================================================
      subroutine ncormat(np,sei,info)
!  =====================================================================
!
!  NCORMAT obtains a positive definite correlation matrix from a
!  symmetric NNCxNNC matrix, whose lower triangle is stored column-wise
!  in vector C.
!
!  INPUT & OUTPUT ARGUMENTS
!  
!  [integer,out]
!  NP   : number of nonpositive eigenvlues of original matrix
!  INFO : 
!         =  0 : successful exit
!         = -1 : iteration failed to converge
!         >  0 : computation of eigenvalues was not successful
!  [double precision,out]
!  SEI  : If NP>0, then the smallest eigenvalue.
!
!  REFERENCES
!
!    [1] N.J. Higham, Computing the nearest correlation matrix, IMA
!        Journal of Numerical Analysis, 22 (2002), 329-343.
!    [2] http://www.nehu-economics.info/nearest-correlation-matrix.html
!
!  ROUTINES AND FUNCTIONS FROM LIBRARIES LAPACK AND BLAS:
!    DLAMCH,DSYEVR
!
!  =====================================================================
!  INPUT & OUTPUT ARGUMENTS
      INTEGER,          INTENT(OUT)   :: NP,INFO
      DOUBLE PRECISION, INTENT(OUT)   :: SEI
!  COUNTERS
      INTEGER          :: I,IT,J,L
!  LOCAL VARIABLES
      INTEGER          :: K,LWORK,LIWORK,M,NIT
      DOUBLE PRECISION :: ACC,DLAMCH,E0,EST,S,SS,X
      PARAMETER(ACC=1.D-10,E0=1.0D-11,NIT=10000)
      DOUBLE PRECISION :: D(NNC,NNC),R(NNC,NNC),V(NNC,NNC),W(NNC)
      LOGICAL          :: TEST
      INTEGER,ALLOCATABLE,DIMENSION(:)          :: IWORK
      DOUBLE PRECISION,ALLOCATABLE,DIMENSION(:) :: WORK
!  EXTERNAL ROUTINES AND FUNCTIONS
      EXTERNAL         :: DLAMCH,DSYEVR
!  =====================================================================
!  INITIALIZATION
      INFO=0
      K=0
      SEI=ZERO
!
!  Save original matrix into R.
!
      DO 5 J=1,NNC
      DO 4 I=J,NNC
      K=K+1
      R(I,J)=C(K)
4     CONTINUE
5     CONTINUE
!
!  NORM VERY LARGE TO BEGIN WITH
!
      S=1.D30
!
!  ITERATION
!
      DO 35 IT=1,NIT
!
!  FIND NONPOSITIVE EIGENVALUES AND CORRESPONDING EIGENVECTORS OF R
!
      LWORK=-1
      LIWORK=-1
!
!  COMPUTE ESTIMATE FOR LARGEST ABSOLUTE VALUE OF EIGENVALUES OF R
!
      DO 8 J=1,NNC
      W(J)=ABS(R(J,J))
8     CONTINUE
      EST=DBLE(NNC-1)+MAXVAL(W)+ACC
      DO 10 J=0,1
      ALLOCATE(WORK(ABS(LWORK)),IWORK(ABS(LIWORK)))
      CALL DSYEVR('V','V','L',NNC,R,NNC,-EST,ZERO,0,0,DLAMCH('S'),M,W,V,
     &NNC,0,WORK,LWORK,IWORK,LIWORK,INFO)
      IF(INFO.GT.0) RETURN
      LWORK=INT(WORK(1))
      LIWORK=IWORK(1)
      DEALLOCATE(WORK,IWORK)
10    CONTINUE
      IF(IT.GT.1) GOTO 12
      NP=M
      IF(M.GT.0) SEI=W(1)
!
!  CONSTRUCT DYKSTRA'S CORRECTION
!
12    DO 14 J=1,NNC
      DO 13 I=J,NNC
      D(I,J)=ZERO
13    CONTINUE
14    CONTINUE
      IF(M.EQ.0) GOTO 18
      DO 17 J=1,NNC
      DO 16 L=1,M
      X=(E0-W(L))*V(J,L)
      DO 15 I=J,NNC
      D(I,J)=D(I,J)+X*V(I,L)
15    CONTINUE
16    CONTINUE
17    CONTINUE
18    K=0
      DO 20 J=1,NNC
      R(J,J)=ONE-D(J,J)
      DO 19 I=J,NNC
      K=K+1
      IF(I.NE.J) R(I,J)=C(K)
19    CONTINUE
20    CONTINUE
!
!  COMPUTE NORM OF DIFFERENCE BETWEEN NEW APPROXIMATION AND ORIGINAL 
!  MATRIX
!
      SS=ZERO
      TEST=.TRUE.
      K=0
      DO 26 J=1,NNC
      DO 24 I=J,NNC
      K=K+1
      IF(I.EQ.J) GOTO 22
!
!  CHECK IF ABSOLUTE VALUES OF NEW CORRELATION COEFFICIENTS ARE GREATER
!  THAN 1 WITH TOLERANCE ACC
!
      IF(ABS(C(K)+D(I,J)).GT.ONE+ACC) TEST=.FALSE.
      SS=SS+2*D(I,J)**2
      GOTO 24
22    SS=SS+(ONE-C(K))**2
24    CONTINUE
26    CONTINUE
      SS=SQRT(SS)
!
!  IF MATRICES HAVE CONVERGED AND LIMIT HAS ENTRIES BETWEEN -1 AND 1,
!  THEN STOP ITERATION
!
      IF(ABS(S-SS).LT.ACC .AND. TEST) GOTO 40
      S=SS
35    CONTINUE
      INFO=-1
      RETURN
40    K=0
      DO 47 J=1,NNC
      DO 45 I=J,NNC
      K=K+1
      IF(I.EQ.J) GOTO 42
      C(K)=C(K)+D(I,J)
      IF(ABS(C(K)).GE.ONE) C(K)=SIGN(ONE-ACC,C(K))
      GOTO 45
42    C(K)=ONE
45    CONTINUE
47    CONTINUE
      RETURN
!  =====================================================================
      end subroutine ncormat
!  =====================================================================
!  =====================================================================
      subroutine file2(sav)
!  =====================================================================
!
!  FILE2 reads resonance parameters from file 2 in ENDF file.
!
!  INPUT & OUTPUT ARGUMENTS
!
!    [logical]
!       SAV :
!         = false: count resonance parameters
!         = true:  save resonance parameters
!
!  =====================================================================
!  INPUT & OUTPUT ARGUMENTS
      logical, intent(in) :: sav
!  COUNTERS
      integer :: i,j
!  VARIABLES
      integer :: l1,l2,n1,n2,njs,nls
!  =====================================================================
      nr=0
      if(.not.sav) goto 5
!
!  Before saving resonance parameters rewind ENDF file to the beginning
!  of energy range records.
!
      do 4 i=1,rec2
      backspace ufil
4     continue
5     rec2=0
      if(lru2.ne.0) goto 7
!
!  If LRU = 0, then one record without resonance parameters is given.
!
      call readl(ufil,rec2)
      return
7     if(lru2.eq.2) goto 40
!
!  LRU = 1
!
      if(lrf2.eq.7) goto 15
!
!  LRU = 1 -> LRF = 1 or 2 or 3
!
      if(nro2.eq.1) call tab1
      call cont(ufil,l1,l2,nls,n2,rec2)
      m2=6
      do 12 i=1,nls
      call cont(ufil,l1,l2,n1,n2,rec2)
!
!  Read/write resonance parameters.
!
      call rwrp(sav,6,n2)
12    continue
      return
!
!  LRU = 1 -> LRF = 7 (only for KBK=0,KPS=0,KRM=3)
!
15    call cont(ufil,l1,l2,njs,n2,rec2)
      m2=1
      if(l2.eq.3) goto 17
      if(l2.ge.1 .and. l2.le.4) call flaimpl('KRM',l2,2)
      call invfla('KRM',l2,2)
17    call listr(ufil,l1,l2,n1,n2,rec2)
      do 30 j=1,njs
      call listr(ufil,l1,l2,n1,n2,rec2)
      if(n2.le.0) call invfla('NCH',n2,2)
!
!  Set M2 to the maximal value of 1+NCH over all NCH's.
!
      m2=max(m2,1+n2)
      if(l1.eq.0 .and. l2.eq.0) goto 21
      if(l1.lt.0) call invfla('KBK',l1,2)
      if(l2.lt.0) call invfla('KPS',l2,2)
      if(l2.gt.0) call flaimpl('KPS',l2,2)
      call flaimpl('KBK',l1,2)
21    call cont(ufil,l1,l2,n1,n1,rec2)
!
!  Read/write resonance parameters.
!
      call rwrp(sav,1+n2,l2)
30    continue
      return
!
!  LRU = 2
!
40    if(lrf2.eq.1) call flaimpl('LRF',lrf2,2)
      if(lrf2.ne.2) call invfla('LRF',lrf2,2)
      if(nro2.eq.1) call tab1
!
!  LRU = 2 -> LRF = 2 (Case C)
!
      m2=6
      call cont(ufil,l1,l2,nls,n2,rec2)
      do 72 i=1,nls
      call cont(ufil,l1,l2,njs,n2,rec2)
      do 70 j=1,njs
      call cont(ufil,l1,l2,n1,n2,rec2)
!
!  Read/write resonance parameters.
!
      call rwrp(sav,6,n2)
70    continue
72    continue
      return
!  =====================================================================
      end subroutine file2
!  =====================================================================
!  =====================================================================
      subroutine rwrp(sav,m,nrs)
!  =====================================================================
!
!  RWRP reads/writes resonance parameters from/to ENDF format.
!
!  =====================================================================
!  INPUT & OUTPUT ARGUMENTS
      logical, intent(in) :: sav
      integer, intent(in) :: m,nrs
!  COUNTERS
      integer             :: i,j,k
!  LOCAL VARIABLES
      integer(kind=1)     :: ll
      double precision    :: a,r(m)
      character(len=10)   :: for
      character(len=12)   :: parr
!  =====================================================================
      if(lru2.eq.2 .and. lrf2.eq.2) call readl(ufil,rec2)
      do 30 i=1,nrs
      nr=nr+1
      if(wri) goto 15
!
!  If WRI=.FALSE., then:
!  - if SAV=.FALSE.: only count resonances and records,
!  - if SAV=.TRUE.: save resonance parameters as well.
!
      read(ufil,for6) (r(j),j=1,m)
      rec2=rec2+1+(m-1)/6
      if(sav) mu(1:m,nr)=r(1:m)
      goto 30
!
!  If WRI=.TRUE., then write values of resonance parameters to FILE2.
!
15    j=0
      do while(j.lt.m)
      read(ufil,fora6) cols,mat,mf,mt,ns
      rec2=rec2+1
      do 20 k=1,6
      if(j.eq.m) goto 23
      j=j+1
      if(.not.sav) goto 20
!
!  If SAV=.TRUE., then write sampled values.
!  Compute length of exponent and set format accordingly.
!
      for='(es12.6e1)'
      ll=9
      if(mu(j,nr).eq.zero) goto 17
      a=abs(mu(j,nr))
      if(a.ge.9.9999995D-10 .and. a.lt.9.9999995D+9) goto 17
      for='(es12.5e2)'
      ll=8
      if(a.ge.9.999995D-100 .and. a.lt.9.999995D+99) goto 17
      for='(es12.4e3)'
      ll=7
17    write(parr,for) mu(j,nr)
!
!  Eliminate 'E'.
!
      write(cols(11*k-10:11*k),fora) parr(1:ll)//parr(ll+2:12)
20    continue
23    write(ufilo,fora6) cols,mat,mf,mt,ns
      end do
30    continue
!  =====================================================================
      end subroutine rwrp
!  =====================================================================
!  =====================================================================
      subroutine tab1
!  =====================================================================
!
!  TAB1 reads a TAB1 record from ENDF format.
!
!  =====================================================================
!  COUNTERS
      integer :: i
!  LOCAL VARIABLES
      integer :: l1,l2,n1,n2
!  =====================================================================
      call cont(ufil,l1,l2,n1,n2,rec2)
      do 5 i=1,2+(n1-1)/3+(n2-1)/3
      call readl(ufil,rec2)
5     continue
      return
!  =====================================================================
      end subroutine tab1
!  =====================================================================
!  =====================================================================
      end module sampling
!  =====================================================================
!  =====================================================================
      program endsam
!  =====================================================================
!
!  ENDSAM reads resonance parameters from a file in ENDF-6 format, makes
!  their samples and writes each sample to a new ENDF file.
!  It reads input parameters from file 'input'. The content of this file
!  must be of the form
!     n
!     file_1
!     ...
!     file_k,
!  where n is a number of samples and file_1,...,file_k are names of
!  ENDF files to be processed.
!  If n is set to 0, then samples are not made, but the information
!  about consistency of covariance data is given.
!
!  =====================================================================
!  MISSING IMPLEMENTATIONS:
!  - NIS>1
!  - LRU=1 -> LRF=4
!  - MF=2 -> LRU=1 -> LRF=7 -> KBK!=0 or KPS!=0 or KRM!=3
!  - MF=2 -> LRU=2 -> LRF=1
!  - MF=32 -> LRU=1 -> NRO=1
!  - MF=32 -> LRU = 1 -> LRF=1or2or3 -> LCOMP = 1 -> NLRS>0
!  =====================================================================
      use endf
      use sampling
      implicit none
!  COUNTERS
      integer            :: i,j,jer2
!  VARIABLES
      integer            :: ii,k,l,nis,nis2,ner2,unin
      logical            :: f32
      double precision   :: a,b,el,eh
      character(len=150) :: endff
      data unin/1/
!  =====================================================================
!
!  Read file with input data.
!
      open(unin,file='input',status='old',action='read')
!
!  Read number of samples.
!
      read(unin,*) n
      if(n.lt.0) goto 777
!
!  Read name of ENDF file.
!
10    read(unin,fora,end=666) endff
      write(*,fora) ' '
      write(*,fora) ' Reading file '//trim(endff)//'.'
!
!  Open ENDF file.
!
      open(ufil,file=endff,status='old',err=999,action='read',
     &blank='zero')
!
!  Initialization.
!  Flag F32 checks whether FILE32 is included in endf file.
!  JER2 is the counter of energy ranges in FILE2.
!
      wri=.false.
      f32=.false.
      jer2=0
!
!  Read a line from endf file.
!
15    read(ufil,fora6,end=145) cols,mat,mf,mt,ns
      if(wri) goto 35
      if(mf.ne.32) goto 15
      if(f32) goto 25
      f32=.true.
!
!  If FILE32 exists, then create output files and temporary file TEMP.
!
      open(utem,file='temp',status='replace',blank='zero')
      do 20 j=1,n
      call chext(endff,j,ii)
      if(ii.gt.0) goto 888
20    continue
25    if(mt.ne.151) goto 30
!
!  Write content of FILE32 to temporary file.
!
      write(utem,fora6) cols,mat,mf,mt,ns
      goto 15
!
!  When reading of FILE32 is finished, rewind ENDF and temporary file
!  and start writing data to output files.
!
30    if(mt.ne.0) call invfla('MT',mt,32)
      wri=.true.
      rewind ufil
      rewind utem
      goto 15
35    do 40 j=1,n
      call chext(endff,j,ii)
40    continue
!
!  Copy records which are not in FILE2 and those records from FILE2
!  which have MT!=151 to output files. Moreover, copy reminder of FILE2
!  if all energy ranges in FILE32 have already been found in FILE2.
!
      if(mf.ne.2 .or. mt.ne.151 .or. jer2.eq.-1) goto 15
!
!  FILE2 & FILE32
!
!  Head in FILE2.
!
      read(cols(45:55),911) nis2
!
!  Head in FILE32.
!
      read(utem,fora6) cols,mat,mf,mt,ns
      read(cols(45:55),911) nis
!
!  Number of isotopes should not be greater in FILE32 than in FILE2.
!  Only the value 1 is implemented.
!
      if(nis.lt.1 .or. nis.gt.nis2) call invfla('NIS',nis,32)
      if(nis2.gt.10) call invfla('NIS',nis2,2)
      if(nis.gt.1) call flaimpl('NIS',nis,32)
      if(nis2.gt.1) call flaimpl('NIS',nis2,2)
!
!  Isotope in FILE2.
!
!  Read energy range record from FILE2.
!
      read(ufil,fora6) cols,mat,mf,mt,ns
      read(cols(34:55),922) lfw,ner2
      if(lfw.ne.0 .and. lfw.ne.1) call invfla('LFW',lfw,2)
!
!  Write energy range record to output files.
!
      do 45 j=1,n
      write(j+9,fora6) cols,mat,mf,mt,ns
45    continue
!
!  Read energy range record from FILE32.
!
      read(utem,923) a,b,k,lfw,ner
      if(lfw.ne.0 .and. lfw.ne.1) call invfla('LFW',lfw,32)
      if(ner.lt.1 .or. ner.gt.ner2) call invfla('NER',ner,32)
      do 140 jer=1,ner
!
!  Energy range in FILE32.
!
      read(utem,923) el,eh,lru,lrf,nro
      if(el.le.zero .or. eh.le.el) goto 733
      if(nro.ne.0 .and. nro.ne.1) call invfla('NRO',nro,32)
      if(lru.eq.1) goto 50
      if(lru.ne.2) call invfla('LRU',lru,32)
      if(lrf.ne.1) call invfla('LRF',lrf,32)
      goto 55
50    if((lrf.ge.1 .and. lrf.le.3) .or. lrf.eq.7) goto 55
      if(lrf.eq.4) call flaimpl('LRF',lrf,32)
      call invfla('LRF',lrf,32)
55    if(jer2.ge.ner2) goto 755
      jer2=jer2+1
!
!  Find energy range among energy ranges in FILE2.
!
      read(ufil,fora6) cols,mat,mf,mt,ns
      do 60 j=1,n
      write(j+9,fora6) cols,mat,mf,mt,ns
60    continue
      read(cols,923) a,b,lru2,lrf2,nro2
      if(a.le.zero .or. b.le.a) goto 733
      if(nro2.ne.0 .and. nro2.ne.1) call invfla('NRO',nro2,2)
      a=(a-el)/min(el,a)
      if(a.gt.-eps) goto 75
!
!  If the current ER from FILE2 is not the one we are looking for, then
!  copy its content into output files.
!
      wri=.true.
      do 70 j=1,n
      ufilo=j+9
      call file2(.false.)
      if(j.eq.n) goto 70
      do 65 i=1,rec2
      backspace ufil
65    continue
70    continue
      goto 55        
75    if(a.ge.eps) goto 755
      if(lru2.ne.lru .or. (lru.eq.1 .and. lrf2.ne.lrf)) goto 766
!
!  If the current ER from FILE2 is the one we are looking for, then
!  start reding resnance parameters from FILE2 and covariance data from
!  FILE32.
!
      if((eh-b)/min(eh,b).lt.eps) goto 80
      write(ou,fori) jer
      write(ov,fori) jer2
      write(*,fora) ' WARNING: Energy ranges '//trim(adjustl(ou))//
     &' from FILE32 and '//trim(adjustl(ov))//' from FILE2 have the same
     & value of EL, but the value of EH is greater in FILE32.'
80    wri=.false.
!
!  Count resonance parameters in FILE2.
!
      call file2(.false.)
      write(*,fora) ' '
      write(ou,fori) nr
      write(*,fora) ' There are '//trim(adjustl(ou))//' resonances in FI
     &LE2.'
      allocate(mu(m2,nr))
!
!  Read resonance parameters from FILE2.
!
      call file2(.true.)
!
!  Count resonance parameters in FILE32.
!
      call file32(.false.,ii)
      write(ou,fori) nnn
      write(ov,fori) nnc
      write(*,fora) ' '
      write(*,fora) ' There are '//trim(adjustl(ou))//' resonance param'
     &//'eters with covariance data. Out of them, there are '
     &//trim(adjustl(ov))//' nonconstant parameters.'
!
!  If there are no nonconstant resonance parameters with covariance
!  data, then proceed to the next energy range.
!
      if(nnc.eq.0) goto 140
      allocate(munc(nnc),nonc(nnc),dev(nnc),cor(nnc*(nnc+1)/2),
     &types(nnc))
!
!  Initialize vectors of covariance data.
!
      munc=zero
      dev=zero
      k=0
      do 95 j=1,nnc
      do 90 i=j,nnc
      k=k+1
      if(i.eq.j) goto 85
      cor(k)=zero
      goto 90
85    cor(k)=one
90    continue
95    continue
      if(lru.eq.2) goto 100
      allocate(ind32(nnn))
      goto 105
100   allocate(ind32(5*nnn))
!
!  Read covariance data from FILE32.
!
105   call file32(.true.,ii)
      write(*,fora) ' '
      if(ii.ne.0) goto 700
      if(nzc.gt.0) goto 110
      write(*,fora) ' Correlation matrix equals identity matrix.'
      goto 115
110   write(ou,fori) nzc
      write(*,fora) ' Correlation matrix contains '//trim(adjustl(ou))//
     &' nonzero nondiagonal correlation(s).'
!
!  Make random samples.
!
115   write(*,fora) ' '
      write(*,fora) ' Making samples.'
      wri=.true.
      allocate(logn(max(1,nnc*(nnc-1)/2)))
      call log_nor_sample(ii)
      deallocate(mu,munc,dev,cor,types,ind32)
      wri=.false.
      write(*,fora) ' '
      write(*,fora) ' Sampling completed.'
      if(ii.eq.0) goto 135
!
!  Give information about inconsistency of covariance data with
!  lognormal distribution.
!
      write(*,fora) ' '
      write(ou,fori) ii
      write(*,fora) ' WARNING: correlations of the following '//
     &trim(adjustl(ou))//' pair(s) of parameters do not satisfy theoreti
     &cal bounds for lognormal distribution:'
      do 130 l=1,ii
      k=logn(l)
      do 125 j=1,nnc
      do 120 i=j,nnc
      k=k-1
      if(k.gt.0) goto 120
      write(ou,fori) nonc(j)
      write(ov,fori) nonc(i)
      write(*,fora) ' ('//trim(adjustl(ou))//','//trim(adjustl(ov))//')'
      goto 130
120   continue
125   continue
130   continue
135   deallocate(logn,nonc)
140   continue
!
!  Set counter of energy ranges in FILE2 to -1 so that remaning ranges
!  will be copied into output files..
!
      jer2=-1
      wri=.true.
      goto 15
!
!  Close input, output, and temporary files.
!
145   close(ufil)
      if(.not.f32) goto 10
      close(utem,status='delete')
      do 150 j=1,n
      inquire(j+9,opened=f32)
      if(f32) close(j+9)
150   continue
      goto 10
!  NORMAL END
666   close(unin)
      write(*,fora) ' '
      stop ' End of process.'
!  =====================================================================
!  ERRORS
700   deallocate(mu,munc,nonc,dev,cor,types,ind32)
      k=0
      do 706 j=1,nnn
      do 705 i=j,nnn
      k=k+1
      if(k.eq.ii) goto 707
705   continue
706   continue
707   write(ou,fori) j
      write(ov,fori) i
      write(*,fora) ' '
      write(*,fora) ' ERROR: absolute value of correlation between pa'
     &//'rameters '//trim(adjustl(ou))//' and '//trim(adjustl(ov))
     &//' is greater than 1.'
      goto 145
!  ---------------------------------------------------------------------
733   write(*,fora) ' ERROR: we must have 0<EL<EH in energy range.'
      goto 145
!  ---------------------------------------------------------------------
755   write(ou,fori) jer
      write(*,fora) ' ERROR: energy range '//trim(adjustl(ou))//
     &' from FILE32 was not found in FILE2.'
      goto 145
!  ---------------------------------------------------------------------
766   write(*,fora) ' ERROR: corresponding energy ranges from FILE2 and 
     &FILE32 must have the same value of LRU. If LRU=1, then the values 
     &of LRF must be the same as well.'
      goto 145
!  ---------------------------------------------------------------------
777   stop ' Number of samples must not be negative.'
!  ---------------------------------------------------------------------
888   write(ou,fori) ii
      write(*,fora) ' ERROR: ENDF file name is '//trim(adjustl(ou))//' c
     &haracters too long.'
      goto 145
!  ---------------------------------------------------------------------
999   write(*,fora) ' ERROR: ENDF file not found.'
      goto 10
!  =====================================================================
!  FORMATS
911   format(i11)
922   format(2i11)
923   format(2e11.0,3i11)
!  =====================================================================
      end program endsam